# -*- coding: utf-8 -*-
# Generated by Django 1.9.7 on 2016-09-01 06:18
from __future__ import unicode_literals

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('sample', '0002_howitworks'),
    ]

    operations = [
        migrations.AddField(
            model_name='product',
            name='product_type',
            field=models.CharField(choices=[('s', 'Sellable'), ('r', 'Rentable')], default='s', max_length=1),
        ),
    ]



"""
Django settings for DRF Auto-endpoint project.

For more information on this file, see
https://docs.djangoproject.com/en/1.9/topics/settings/

For the full list of settings and their values, see
https://docs.djangoproject.com/en/1.9/ref/settings/
"""

# Build paths inside the project like this: os.path.join(BASE_DIR, ...)
import os

BASE_DIR = os.path.dirname(os.path.abspath(__file__))


# Quick-start development settings - unsuitable for production
# See https://docs.djangoproject.com/en/1.9/howto/deployment/checklist/

try:
    from environ import *
except ImportError:
    raise Exception('Improperly configured: Please make an untracked copy of '
                    'environ.py.dist to environ.py and update with your '
                    'actual settings')

# Application definition

INSTALLED_APPS = (
    'sample',

    'django.contrib.admin',
    'django.contrib.auth',
    'django.contrib.contenttypes',
    'django.contrib.sessions',
    'django.contrib.messages',
    'django.contrib.staticfiles',

    'rest_framework',

    'drf_auto_endpoint',
)

MIDDLEWARE_CLASSES = (
    'django.contrib.sessions.middleware.SessionMiddleware',
    'django.middleware.common.CommonMiddleware',
    'django.middleware.csrf.CsrfViewMiddleware',
    'django.contrib.auth.middleware.AuthenticationMiddleware',
    'django.contrib.auth.middleware.SessionAuthenticationMiddleware',
    'django.contrib.messages.middleware.MessageMiddleware',
    'django.middleware.clickjacking.XFrameOptionsMiddleware',
    'django.middleware.security.SecurityMiddleware',
)

ROOT_URLCONF = 'urls'

TEMPLATES = [
    {
        'BACKEND': 'django.template.backends.django.DjangoTemplates',
        'DIRS': [],
        'APP_DIRS': True,
        'OPTIONS': {
            'context_processors': [
                'django.template.context_processors.debug',
                'django.template.context_processors.request',
                'django.contrib.auth.context_processors.auth',
                'django.contrib.messages.context_processors.messages',
            ],
        },
    },
]

WSGI_APPLICATION = 'sample.wsgi.application'


# Database
# https://docs.djangoproject.com/en/1.9/ref/settings/#databases

DATABASES = {
    'default': {
        'ENGINE': 'django.db.backends.sqlite3',
        'NAME': os.path.join(BASE_DIR, 'db.sqlite3'),
    }
}


# Internationalization
# https://docs.djangoproject.com/en/1.9/topics/i18n/

LANGUAGE_CODE = 'en-us'

TIME_ZONE = 'UTC'

USE_I18N = True

USE_L10N = True

USE_TZ = True


# Static files (CSS, JavaScript, Images)
# https://docs.djangoproject.com/en/1.9/howto/static-files/

STATIC_URL = '/static/'
STATIC_ROOT = os.path.join(BASE_DIR, 'static')

REST_FRAMEWORK = {
    'DEFAULT_METADATA_CLASS': 'drf_auto_endpoint.metadata.AutoMetadata',
}



from rest_framework import viewsets, serializers

from ..models import Product


class DummyProductSerializer(serializers.ModelSerializer):

    class Meta:
        model = Product
        fields = ('id', 'name', 'category')


class DummyProductViewSet(viewsets.ModelViewSet):

    serializer_class = DummyProductSerializer
    queryset = Product.objects.all()



# -*- coding: utf-8 -*-
# Generated by Django 1.9.7 on 2016-06-22 04:20
from __future__ import unicode_literals

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    initial = True

    dependencies = [
    ]

    operations = [
        migrations.CreateModel(
            name='Category',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('name', models.CharField(max_length=255)),
            ],
            options={
                'verbose_name_plural': 'categories',
            },
        ),
        migrations.CreateModel(
            name='Product',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('name', models.CharField(max_length=255)),
                ('category', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='products', to='sample.Category')),
            ],
        ),
    ]



from django.db import models

from django.utils.translation import ugettext_lazy as _


PRODUCT_TYPES = (
    ('s', 'Sellable'),
    ('r', 'Rentable'),
)


class Category(models.Model):

    name = models.CharField(max_length=255)

    class Meta:
        verbose_name_plural = 'categories'

    def __str__(self):
        return self.name


class Product(models.Model):

    name = models.CharField(max_length=255)
    category = models.ForeignKey(Category, related_name='products')
    product_type = models.CharField(max_length=1, choices=PRODUCT_TYPES, default='s')

    def __str__(self):
        return self.name

class HowItWorks(models.Model):

    name = models.CharField(max_length=255)

    class Meta:
        verbose_name = _('How it works')
        verbose_name_plural = _('How it works')

    def __str__(self):
        return self.name



from rest_framework.test import APITestCase

from .factories import CategoryFactory, ProductFactory, HowItWorksFactory
from .base import EndpointAPITestCase

from ..views import AbstractHowItWorksViewSet
from urls import router


class CategoryAPITest(EndpointAPITestCase, APITestCase):

    endpoint_url = 'sample/categories'
    model_factory_class = CategoryFactory

    create_requires_login = False
    update_requires_login = False
    delete_requires_login = False


class ProductAPITest(EndpointAPITestCase, APITestCase):

    endpoint_url = 'sample/products'
    model_factory_class = ProductFactory

    create_requires_login = False
    update_requires_login = False
    delete_requires_login = False



class HowItWorksAPITest(EndpointAPITestCase, APITestCase):

    endpoint_url = 'sample/howitworks'
    model_factory_class = HowItWorksFactory

    create_requires_login = False
    update_requires_login = False
    delete_requires_login = False

    def setUp(self):
        super(HowItWorksAPITest, self).setUp()
        self.called = router._endpoints[self.endpoint_url].viewset.called_counter

    def test_list_view(self):
        super(HowItWorksAPITest, self).test_list_view()
        self.assertTrue(self.called < router._endpoints[self.endpoint_url].viewset.called_counter,
                        router._endpoints[self.endpoint_url].viewset.called_counter)



from django.conf.urls import include, url
from django.contrib import admin

from drf_auto_endpoint.router import router
from sample.views import HomepageView, AbstractHowItWorksViewSet

from sample.models import Category, Product, HowItWorks

router.register(Category)
router.register(Product)
router.register(HowItWorks, base_viewset=AbstractHowItWorksViewSet)

urlpatterns = [
    # Examples:
    # url(r'^$', 'drf_auto_endpoint.views.home', name='home'),
    # url(r'^blog/', include('blog.urls')),

    url(r'^admin/', include(admin.site.urls)),
    url(r'^$', HomepageView.as_view()),
    url(r'^api/', include(router.urls)),
]



from rest_framework.metadata import SimpleMetadata, BaseMetadata


class AutoMetadataMixin:

    def determine_metadata(self, request, view):
        from drf_auto_endpoint.app_settings import settings

        try:
            metadata = super(AutoMetadataMixin, self).determine_metadata(request, view)
        except NotImplementedError:
            metadata = {}

        if not hasattr(view, 'endpoint'):
            serializer_instance = view.serializer_class()
            metadata['fields'] = []
            for field in view.serializer_class.Meta.fields:
                instance_field = serializer_instance.fields[field]
                if field == 'id':
                    continue
                widget = None
                extra = {}
                if getattr(instance_field, 'choices', None) is not None:
                    widget = settings.WIDGET_MAPPING['choice']
                    extra['choices'] = [
                        {
                            'label': v,
                            'value': k,
                        } for k, v in instance_field.choices.items()
                    ]
                metadata['fields'].append({
                    'name': field,
                    'label': field.title(),
                    'widget': widget if widget is not None else \
                        settings.WIDGET_MAPPING[instance_field.__class__.__name__],
                    'extra': extra
                })
            metadata.update({
                'list_display': ['__str__', ],
                'filter_fields': [],
                'search_fields': [],
                'ordering_fields': [],
                'fieldsets': [{'title': None, 'fields': [
                    field
                    for field in view.serializer_class.Meta.fields
                    if field != 'id' and field != '__str__'
                ]}]
            })
            return metadata

        for prop in ['fields', 'list_display', 'filter_fields', 'search_enabled', 'ordering_fields',
                     'needs', 'fieldsets']:
            metadata[prop] = getattr(view.endpoint, 'get_{}'.format(prop))()

        return metadata


class AutoMetadata(AutoMetadataMixin, SimpleMetadata):
    pass


class MinimalAutoMetadata(AutoMetadataMixin, BaseMetadata):
    pass



from six import iteritems

from django.core.urlresolvers import reverse

from rest_framework import status

from .factories import get_admin


class BaseAPITestCase(object):
    model_factory_class = None
    model = None
    factory_default_kwargs = {}
    factory_delete_kwargs = {}
    api_base_name = None
    api_base_url = ''

    list_requires_login = False
    detail_requires_login = False
    create_requires_login = True
    update_requires_login = True
    delete_requires_login = True

    to_update_fieds = {}
    extra_update_fields = {}
    create_kwargs = {}
    extra_create_fields = {}
    create_strip_data_fields = ['id', ]
    create_strip_response_fields = ['id', ]

    serializer_class = None
    create_serializer_class = None
    update_serializer_class = None

    api_is_read_only = False

    @classmethod
    def setUpTestData(cls):
        cls.test_model = cls.model_factory_class(
            **cls.factory_default_kwargs
        )
        cls.test_model.save()
        cls.admin = get_admin()

    def tearDown(self):
        self.client.logout()

    def login(self, user=None):
        if user is None:
            user = self.admin

        self.client.force_authenticate(user)

    def get_list_url(self):
        if self.api_base_name is None:
            return self.api_base_url

        return reverse('{}-list'.format(self.api_base_name))

    def get_detail_url(self, record=None):
        if record is None:
            record = self.test_model

        if self.api_base_name is None:
            return '{}{}/'.format(
                 self.api_base_url,
                 self.test_model.pk
            )

        return reverse(
            '{}-detail'.format(self.api_base_name),
            kwargs={'pk': record.pk, }
        )

    def test_list_view(self):
        url = self.get_list_url()
        response = self.client.get(url, format='json')

        if self.list_requires_login:
            self.assertIn(response.status_code, [status.HTTP_403_FORBIDDEN, status.HTTP_401_UNAUTHORIZED])
            self.login()
            response = self.client.get(url, format='json')

        self.assertEqual(response.status_code, status.HTTP_200_OK)

    def test_detail_view(self):
        url = self.get_detail_url()
        response = self.client.get(url, format='json')

        if self.detail_requires_login:
            self.assertIn(response.status_code, [status.HTTP_403_FORBIDDEN, status.HTTP_401_UNAUTHORIZED])
            self.login()
            response = self.client.get(url, format='json')

        self.assertEqual(response.status_code, status.HTTP_200_OK)

    def test_update_view(self):
        if self.api_is_read_only:
            self.assertTrue(True)
        else:
            url = self.get_detail_url()

            record = self.model.objects.get(
                pk=self.test_model.pk
            )

            for field, value in iteritems(self.to_update_fieds):
                setattr(record, field, value)
            if self.update_serializer_class is not None:
                serializer = self.update_serializer_class(record)
            else:
                serializer = self.serializer_class(record)

            data = serializer.data

            for field, value in iteritems(self.extra_update_fields):
                data[field] = value

            response = self.client.put(url, data, format='json')

            if self.update_requires_login:
                self.assertIn(response.status_code, [status.HTTP_403_FORBIDDEN, status.HTTP_401_UNAUTHORIZED])
                self.login()
                response = self.client.put(url, data, format='json')

            self.assertEqual(response.status_code, status.HTTP_200_OK)

    def test_delete_view(self):
        if self.api_is_read_only:
            self.assertTrue(True)
        else:
            to_delete = self.model_factory_class(**self.factory_delete_kwargs)
            to_delete.save()

            url = self.get_detail_url(to_delete)
            response = self.client.delete(url, format='json')

            if self.delete_requires_login:
                self.assertIn(response.status_code, [status.HTTP_403_FORBIDDEN, status.HTTP_401_UNAUTHORIZED])
                self.login()
                response = self.client.delete(url, format='json')

            self.assertEqual(response.status_code, status.HTTP_204_NO_CONTENT)

    def test_create_view(self):
        if self.api_is_read_only:
            self.assertTrue(True)
        else:
            url = self.get_list_url()

            dummy = self.model_factory_class.simple_generate(
                create=False,
                **self.create_kwargs
            )

            foreign_keys = [field.name for field in self.serializer_class.Meta.model._meta.get_fields()
                                         if field.one_to_one or (field.many_to_one and field.related_model)]

            if self.create_serializer_class is not None:
                serializer = self.create_serializer_class(dummy)
            else:
                serializer = self.serializer_class(dummy)

            data = serializer.data

            for field in foreign_keys:
                related = getattr(dummy, field, None)
                if related is not None:
                        related.save()
                        data[field] = related.id

            for field, value in iteritems(self.extra_create_fields):
                data[field] = value

            response = self.client.post(url, data, format='json')

            if self.create_requires_login:
                self.assertIn(response.status_code, [status.HTTP_403_FORBIDDEN, status.HTTP_401_UNAUTHORIZED])
                self.login()
                response = self.client.post(url, data, format='json')

            self.assertEqual(response.status_code, status.HTTP_201_CREATED, response.content)

            response_data = response.data
            for field in self.create_strip_response_fields:
                response_data.pop(field, None)
            for field in self.create_strip_data_fields:
                data.pop(field, None)

            self.assertEqual(response_data, data)


class EndpointAPITestCase(BaseAPITestCase):

    api_url_prefix = '/api/'

    def __init__(self, *args, **kwargs):
        from urls import router
        self.api_base_url =  '{}{}/'.format(self.api_url_prefix, self.endpoint_url)
        self.endpoint = router.get_endpoint(self.endpoint_url)
        self.model = self.endpoint.model
        self.serializer_class = self.endpoint.get_serializer()
        self.api_is_read_only = self.endpoint.read_only

        super(EndpointAPITestCase, self).__init__(*args, **kwargs)



from django.test import TestCase

from rest_framework.permissions import AllowAny
from rest_framework import filters, pagination

from ..models import Product, Category, PRODUCT_TYPES

from .data import DummyProductSerializer, DummyProductViewSet

from drf_auto_endpoint.endpoints import Endpoint
from drf_auto_endpoint.router import router


class EndpointTestCase(TestCase):

    def setUp(self):
        self.endpoint = Endpoint(model=Product)
        self.fields = tuple([field.name for field in Product._meta.get_fields()])

        self.permission_classes = (AllowAny, )
        self.filter_fields = ('name', 'category_id')
        self.search_fields = ('id', 'name')
        self.ordering_fields = ('name', )
        self.page_size = 2

        self.alternate_endpoint = Endpoint(model=Product, permission_classes=self.permission_classes,
                                           filter_fields=self.filter_fields,
                                           search_fields=self.search_fields,
                                           ordering_fields=self.ordering_fields,
                                           page_size=self.page_size)

    def test_fields(self):
        self.assertEqual(len(self.endpoint.get_fields_for_serializer()), len(self.fields) + 1)
        fields_for_serializer = self.endpoint.get_fields_for_serializer()
        for field in self.fields:
            self.assertIn(field, fields_for_serializer)
        self.assertIn('__str__', fields_for_serializer)

        endpoint = Endpoint(model=Product, include_str=False)
        self.assertEqual(len(endpoint.get_fields_for_serializer()), len(self.fields))
        for field in self.fields:
            self.assertIn(field, endpoint.get_fields_for_serializer())
        self.assertNotIn('__str__', endpoint.get_fields_for_serializer())

        endpoint = Endpoint(model=Product, fields=('id', 'name'))
        self.assertEqual(len(endpoint.get_fields_for_serializer()), 2)
        for field in ('id', 'name'):
            self.assertIn(field, endpoint.get_fields_for_serializer())

    def test_widgets(self):

        data = {
            'name': 'text',
            'category': 'foreignkey',
            'product_type': 'select',
        }

        for field, expected_widget in data.items():
            field_dict = self.endpoint._get_field_dict(field)
            self.assertIn('widget', field_dict)
            self.assertEqual(field_dict['widget'], expected_widget)
            self.assertIn('extra', field_dict)
            if field == 'category':
                self.assertIn('related_model', field_dict['extra'])
                self.assertEqual(field_dict['extra']['related_model'], 'sample/category')
                self.assertNotIn('choices', field_dict['extra'])
            elif field == 'product_type':
                self.assertIn('choices', field_dict['extra'])
                self.assertEqual(len(field_dict['extra']['choices']), len(PRODUCT_TYPES))

    def test_model(self):
        self.assertEqual(self.endpoint.model, Product)

    def test_dynamic_params(self):
        endpoint = self.alternate_endpoint

        for attr in ('permission_classes', 'filter_fields', 'search_fields', 'ordering_fields',
                     'page_size'):
            self.assertEqual(getattr(endpoint, attr), getattr(self, attr))

    def test_serializer_factory(self):
        serializer = self.endpoint.get_serializer()
        self.assertEqual(serializer.Meta.model, Product)
        self.assertEqual(len(serializer.Meta.fields), len(self.endpoint.get_fields_for_serializer()))
        for field in self.fields:
            self.assertIn(field, serializer.Meta.fields)

        self.assertEqual(serializer.__name__, 'ProductSerializer')

    def test_non_factory_serializer(self):
        endpoint = Endpoint(model=Product, serializer=DummyProductSerializer)

        self.assertEqual(endpoint.get_serializer(), DummyProductSerializer)

    def test_viewset_factory(self):
        viewset = self.endpoint.get_viewset()
        self.assertEqual(viewset.serializer_class, self.endpoint.get_serializer())

        for backend in (filters.DjangoFilterBackend, filters.SearchFilter):
            self.assertNotIn(backend, viewset.filter_backends)

        self.assertEqual(viewset.__name__, 'ProductViewSet')

        viewset = self.alternate_endpoint.get_viewset()

        for attr in ('permission_classes', 'filter_fields', 'search_fields', 'ordering_fields'):
            self.assertEqual(getattr(viewset, attr), getattr(self, attr))

        for backend in (filters.DjangoFilterBackend, filters.SearchFilter, filters.OrderingFilter):
            self.assertIn(backend, viewset.filter_backends)

        self.assertEqual(viewset.pagination_class, pagination.PageNumberPagination)

    def test_non_factory_viewset(self):
        endpoint = Endpoint(viewset=DummyProductViewSet)

        self.assertEqual(endpoint.get_viewset(), DummyProductViewSet)
        self.assertEqual(endpoint.get_serializer(), DummyProductSerializer)
        self.assertEqual(endpoint.model, Product)

    def test_inheritance(self):

        class DummyEndpoint(Endpoint):
            model = Product
            permission_classes = self.permission_classes
            filter_fields = self.filter_fields
            search_fields = self.search_fields
            ordering_fields = self.ordering_fields

        endpoint = DummyEndpoint()

        self.assertEqual(endpoint.model, Product)
        self.assertEqual(len(endpoint.get_fields_for_serializer()), len(self.fields) + 1)
        for field in self.fields:
            self.assertIn(field, endpoint.get_fields_for_serializer())
        self.assertIn('__str__', endpoint.get_fields_for_serializer())

        serializer = endpoint.get_serializer()
        self.assertEqual(serializer.Meta.model, Product)
        self.assertEqual(len(serializer.Meta.fields), len(self.endpoint.get_fields_for_serializer()))

        viewset = endpoint.get_viewset()

        for attr in ('permission_classes', 'filter_fields', 'search_fields', 'ordering_fields'):
            self.assertEqual(getattr(viewset, attr), getattr(self, attr))

        for backend in (filters.DjangoFilterBackend, filters.SearchFilter, filters.OrderingFilter):
            self.assertIn(backend, viewset.filter_backends)

    def test_plurals(self):
        endpoint = Endpoint(model=Category)
        self.assertEqual(endpoint.get_url(), 'sample/categories')

        needs = self.endpoint.get_needs()
        self.assertEqual(len(needs), 1)
        self.assertEqual(needs[0]['plural'], 'categories')


class RouterTestCase(TestCase):

    def test_register_model(self):
        router.register(Product)
        endpoint = router.get_endpoint('sample/products')
        self.assertTrue(isinstance(endpoint, Endpoint))

    def test_register_endpoint(self):
        router.register(endpoint=Endpoint(Product))
        endpoint = router.get_endpoint('sample/products')
        self.assertTrue(isinstance(endpoint, Endpoint))

    def test_alternate_url(self):
        router.register(Product, url='bogus')
        endpoint = router.get_endpoint('bogus')
        self.assertTrue(isinstance(endpoint, Endpoint))



"""
WSGI config for DRF Auto-endpoint project.

It exposes the WSGI callable as a module-level variable named ``application``.

For more information on this file, see
https://docs.djangoproject.com/en/1.9/howto/deployment/wsgi/
"""

import os
import sys
from django.core.wsgi import get_wsgi_application


BASE_DIR = os.path.dirname(os.path.dirname(os.path.abspath(__file__)))

os.environ["DJANGO_SETTINGS_MODULE"] = "settings"
sys.path.append(BASE_DIR)

application = get_wsgi_application()



from django.conf import settings

from factory import fuzzy, SubFactory, Faker, PostGenerationMethodCall, lazy_attribute
from factory.django import DjangoModelFactory

from ..models import Category, Product, HowItWorks


class UserFactory(DjangoModelFactory):

    class Meta:
        model = settings.AUTH_USER_MODEL

    username = fuzzy.FuzzyText()
    first_name = fuzzy.FuzzyText()
    last_name = fuzzy.FuzzyText()
    password = PostGenerationMethodCall('set_password', 'changeM3')
    is_active = True
    is_staff = False
    is_superuser = False

    @lazy_attribute
    def email(self):
        return '{}@example.com'.format(self.username)


def get_admin(username=None):
    from django.contrib.auth import get_user_model

    kwargs = {
        'is_superuser': True,
        'is_staff': True
    }

    if username is not None:
        kwargs['username'] = username

    users = get_user_model().objects.filter(**kwargs)
    if users.count() > 0:
        return users[0]

    rv = UserFactory(**kwargs)
    rv.save()

    return rv


class CategoryFactory(DjangoModelFactory):

    class Meta:
        model = Category

    name = fuzzy.FuzzyText(prefix='Category ', length=3)


class ProductFactory(DjangoModelFactory):

    class Meta:
        model = Product

    name = Faker('word')
    category = SubFactory(CategoryFactory)


class HowItWorksFactory(DjangoModelFactory):

    class Meta:
        model = HowItWorks

    name = Faker('word')



from django.views.generic import RedirectView

from rest_framework import viewsets


class HomepageView(RedirectView):

    url = '/api/'


class AbstractHowItWorksViewSet(viewsets.ModelViewSet):

    called_counter = 0

    def get_queryset(self):
        self.__class__.called_counter += 1
        return super(AbstractHowItWorksViewSet, self).get_queryset()
